@@ -143,29 +143,19 @@ static void
 gupnp_control_point_dispose (GObject *object)
 {
         GUPnPControlPoint *control_point;
+        GSSDPResourceBrowser *browser;
         GObjectClass *object_class;
 
         control_point = GUPNP_CONTROL_POINT (object);
+        browser = GSSDP_RESOURCE_BROWSER (control_point);
+
+        gssdp_resource_browser_set_active (browser, FALSE);
 
         if (control_point->priv->factory) {
                 g_object_unref (control_point->priv->factory);
                 control_point->priv->factory = NULL;
         }
 
-        while (control_point->priv->devices) {
-                g_object_unref (control_point->priv->devices->data);
-                control_point->priv->devices =
-                        g_list_delete_link (control_point->priv->devices,
-                                            control_point->priv->devices);
-        }
-
-        while (control_point->priv->services) {
-                g_object_unref (control_point->priv->services->data);
-                control_point->priv->services =
-                        g_list_delete_link (control_point->priv->services,
-                                            control_point->priv->services);
-        }
-
         /* Cancel any pending description file GETs */
         while (control_point->priv->pending_gets) {
                 GetDescriptionURLData *data;
@@ -209,21 +199,145 @@ gupnp_control_point_finalize (GObject *object)
         object_class->finalize (object);
 }
 
+static GList *
+find_service_node (GUPnPControlPoint *control_point,
+                   const char        *udn,
+                   const char        *service_type)
+{
+        GList *l;
+
+        l = control_point->priv->services;
+
+        while (l) {
+                GUPnPServiceInfo *info;
+
+                info = GUPNP_SERVICE_INFO (l->data);
+
+                if ((strcmp (gupnp_service_info_get_udn (info), udn) == 0) ||
+                    (strcmp (gupnp_service_info_get_service_type (info),
+                             service_type) == 0))
+                        break;
+
+                l = l->next;
+        }
+
+        return l;
+}
+
+static GList *
+find_device_node (GUPnPControlPoint *control_point,
+                  const char        *udn)
+{
+        GList *l;
+
+        l = control_point->priv->devices;
+
+        while (l) {
+                GUPnPDeviceInfo *info;
+
+                info = GUPNP_DEVICE_INFO (l->data);
+
+                if (strcmp (udn, gupnp_device_info_get_udn (info)) == 0)
+                        break;
+
+                l = l->next;
+        }
+
+        return l;
+}
+
+static void
+create_and_report_service_proxy (GUPnPControlPoint  *control_point,
+                                 GUPnPXMLDoc        *doc,
+                                 xmlNode            *element,
+                                 const char         *udn,
+                                 const char         *service_type,
+                                 const char         *description_url,
+                                 SoupURI            *url_base)
+{
+        GUPnPServiceProxy *proxy;
+        GUPnPResourceFactory *factory;
+        GUPnPContext *context;
+
+        if (find_service_node (control_point, udn, service_type) != NULL)
+                /* We already have a proxy for this service */
+                return;
+
+        factory = gupnp_control_point_get_resource_factory (control_point);
+        context = gupnp_control_point_get_context (control_point);
+
+        /* Create proxy */
+        proxy = gupnp_resource_factory_create_service_proxy (factory,
+                                                             context,
+                                                             doc,
+                                                             element,
+                                                             udn,
+                                                             service_type,
+                                                             description_url,
+                                                             url_base);
+
+        control_point->priv->services =
+                                g_list_prepend (control_point->priv->services,
+                                                proxy);
+
+        g_signal_emit (control_point,
+                       signals[SERVICE_PROXY_AVAILABLE],
+                       0,
+                       proxy);
+}
+
+static void
+create_and_report_device_proxy (GUPnPControlPoint  *control_point,
+                                GUPnPXMLDoc        *doc,
+                                xmlNode            *element,
+                                const char         *udn,
+                                const char         *description_url,
+                                SoupURI            *url_base)
+{
+        GUPnPDeviceProxy *proxy;
+        GUPnPResourceFactory *factory;
+        GUPnPContext *context;
+
+        if (find_device_node (control_point, udn) != NULL)
+                /* We already have a proxy for this device */
+                return;
+
+        factory = gupnp_control_point_get_resource_factory (control_point);
+        context = gupnp_control_point_get_context (control_point);
+
+        proxy = gupnp_resource_factory_create_device_proxy (factory,
+                                                            context,
+                                                            doc,
+                                                            element,
+                                                            udn,
+                                                            description_url,
+                                                            url_base);
+
+        control_point->priv->devices = g_list_prepend
+                                                (control_point->priv->devices,
+                                                 proxy);
+
+        g_signal_emit (control_point,
+                       signals[DEVICE_PROXY_AVAILABLE],
+                       0,
+                       proxy);
+}
+
 /* Search @element for matching services */
 static void
 process_service_list (xmlNode           *element,
                       GUPnPControlPoint *control_point,
-                      XmlDocWrapper     *doc,
+                      GUPnPXMLDoc       *doc,
                       const char        *udn,
                       const char        *service_type,
                       const char        *description_url,
                       SoupURI           *url_base)
 {
+        g_object_ref (control_point);
+
         for (element = element->children; element; element = element->next) {
                 xmlChar *prop;
                 gboolean match;
-                GUPnPContext *context;
-                GUPnPServiceProxy *proxy;
 
                 if (strcmp ((char *) element->name, "service") != 0)
                         continue;
@@ -243,46 +357,34 @@ process_service_list (xmlNode           *element,
 
                 /* Match */
 
-                /* Get context */
-                context = gupnp_control_point_get_context (control_point);
-
-                /* Create proxy */
-                proxy = gupnp_resource_factory_create_service_proxy
-                                                 (control_point->priv->factory,
-                                                  context,
-                                                  doc,
-                                                  element,
-                                                  udn,
-                                                  service_type,
-                                                  description_url,
-                                                  url_base);
-
-                control_point->priv->services =
-                        g_list_prepend (control_point->priv->services,
-                                        proxy);
-
-                g_signal_emit (control_point,
-                               signals[SERVICE_PROXY_AVAILABLE],
-                               0,
-                               proxy);
+                create_and_report_service_proxy (control_point,
+                                                 doc,
+                                                 element,
+                                                 udn,
+                                                 service_type,
+                                                 description_url,
+                                                 url_base);
         }
+
+        g_object_unref (control_point);
 }
 
 /* Recursively search @element for matching devices */
 static void
 process_device_list (xmlNode           *element,
                      GUPnPControlPoint *control_point,
-                     XmlDocWrapper     *doc,
+                     GUPnPXMLDoc       *doc,
                      const char        *udn,
                      const char        *service_type,
                      const char        *description_url,
                      SoupURI           *url_base)
 {
+        g_object_ref (control_point);
+
         for (element = element->children; element; element = element->next) {
                 xmlNode *children;
                 xmlChar *prop;
                 gboolean match;
-                GUPnPContext *context;
 
                 if (strcmp ((char *) element->name, "device") != 0)
                         continue;
@@ -316,9 +418,6 @@ process_device_list (xmlNode           *element,
 
                 /* Match */
 
-                /* Get context */
-                context = gupnp_control_point_get_context (control_point);
-
                 if (service_type) {
                         /* Dive into serviceList */
                         children = xml_util_get_element (element,
@@ -334,30 +433,16 @@ process_device_list (xmlNode           *element,
                                                       description_url,
                                                       url_base);
                         }
-                } else {
-                        /* Create device proxy */
-                        GUPnPDeviceProxy *proxy;
-
-                        proxy = gupnp_resource_factory_create_device_proxy
-                                        (control_point->priv->factory,
-                                         context,
-                                         doc,
-                                         element,
-                                         udn,
-                                         description_url,
-                                         url_base);
-
-                        control_point->priv->devices =
-                                g_list_prepend
-                                        (control_point->priv->devices,
-                                         proxy);
-
-                        g_signal_emit (control_point,
-                                       signals[DEVICE_PROXY_AVAILABLE],
-                                       0,
-                                       proxy);
-                }
+                } else
+                        create_and_report_device_proxy (control_point,
+                                                        doc,
+                                                        element,
+                                                        udn,
+                                                        description_url,
+                                                        url_base);
         }
+
+        g_object_unref (control_point);
 }
 
 /*
@@ -365,7 +450,7 @@ process_device_list (xmlNode           *element,
  */
 static void
 description_loaded (GUPnPControlPoint *control_point,
-                    XmlDocWrapper     *doc,
+                    GUPnPXMLDoc       *doc,
                     const char        *udn,
                     const char        *service_type,
                     const char        *description_url)
@@ -377,6 +462,17 @@ description_loaded (GUPnPControlPoint *control_point,
         element = xml_util_get_element ((xmlNode *) doc->doc,
                                         "root",
                                         NULL);
+        if (element == NULL) {
+                g_warning ("No 'root' element found in description document"
+                           " '%s'. Ignoring device '%s'\n",
+                           description_url,
+                           udn);
+
+                return;
+        }
+
+        if (element == NULL)
+                return;
 
         url_base = xml_util_get_child_element_content_uri (element,
                                                            "URLBase",
@@ -405,7 +501,7 @@ got_description_url (SoupSession           *session,
                      SoupMessage           *msg,
                      GetDescriptionURLData *data)
 {
-        XmlDocWrapper *doc;
+        GUPnPXMLDoc *doc;
 
         if (msg->status_code == SOUP_STATUS_CANCELLED)
                 return;
@@ -432,10 +528,10 @@ got_description_url (SoupSession           *session,
                 xmlDoc *xml_doc;
 
                 /* Parse response */
-                xml_doc = xmlParseMemory (msg->response_body->data,
-                                          msg->response_body->length);
+                xml_doc = xmlRecoverMemory (msg->response_body->data,
+                                            msg->response_body->length);
                 if (xml_doc) {
-                        doc = xml_doc_wrapper_new (xml_doc);
+                        doc = gupnp_xml_doc_new (xml_doc);
 
                         description_loaded (data->control_point,
                                             doc,
@@ -456,12 +552,13 @@ got_description_url (SoupSession           *session,
                                            data->control_point);
 
                         /* If no proxy was created, make sure doc is freed. */
-                        g_object_ref_sink (doc);
                         g_object_unref (doc);
                 } else
                         g_warning ("Failed to parse %s", data->description_url);
         } else
-                g_warning ("Failed to GET %s", data->description_url);
+                g_warning ("Failed to GET %s: %s",
+                           data->description_url,
+                           msg->reason_phrase);
 
         get_description_url_data_free (data);
 }
@@ -480,7 +577,7 @@ load_description (GUPnPControlPoint *control_point,
                   const char        *udn,
                   const char        *service_type)
 {
-        XmlDocWrapper *doc;
+        GUPnPXMLDoc *doc;
 
         doc = g_hash_table_lookup (control_point->priv->doc_cache,
                                    description_url);
@@ -559,7 +656,7 @@ parse_usn (const char *usn,
         bits = g_strsplit (usn, "::", -1);
 
         /* Count elements */
-        for (count = 0; bits[count]; count++);
+        count = g_strv_length (bits);
 
         if (count == 1) {
                 /* uuid:device-UUID */
@@ -570,17 +667,21 @@ parse_usn (const char *usn,
 
         } else if (count == 2) {
                 char **second_bits;
+                guint n_second_bits;
 
                 second_bits = g_strsplit (bits[1], ":", -1);
+                n_second_bits = g_strv_length (second_bits);
 
-                if (!strcmp (second_bits[0], "upnp") &&
+                if (n_second_bits >= 2 &&
+                    !strcmp (second_bits[0], "upnp") &&
                     !strcmp (second_bits[1], "rootdevice")) {
                         /* uuid:device-UUID::upnp:rootdevice */
 
                         *udn = bits[0];
 
                         ret = TRUE;
-                } else if (!strcmp (second_bits[0], "urn")) {
+                } else if (n_second_bits >= 3 &&
+                           !strcmp (second_bits[0], "urn")) {
                         /* uuid:device-UIID::urn:domain-name:service/device:
                          * type:v */
 
@@ -659,26 +760,13 @@ gupnp_control_point_resource_unavailable
 
         /* Find proxy */
         if (service_type) {
-                l = control_point->priv->services;
+                l = find_service_node (control_point, udn, service_type);
 
-                while (l) {
-                        GUPnPServiceInfo *info;
+                if (l) {
                         GUPnPServiceProxy *proxy;
 
-                        info = GUPNP_SERVICE_INFO (l->data);
-
-                        if ((strcmp (udn,
-                                     gupnp_service_info_get_udn (info)) != 0) ||
-                            (strcmp (service_type,
-                                     gupnp_service_info_get_service_type (info))
-                                     != 0)) {
-                                l = l->next;
-
-                                continue;
-                        }
-
                         /* Remove proxy */
-                        proxy = GUPNP_SERVICE_PROXY (info);
+                        proxy = GUPNP_SERVICE_PROXY (l->data);
 
                         cur_l = l;
                         l = l->next;
@@ -695,23 +783,13 @@ gupnp_control_point_resource_unavailable
                         g_object_unref (proxy);
                 }
         } else {
-                l = control_point->priv->devices;
+                l = find_device_node (control_point, udn);
 
-                while (l) {
-                        GUPnPDeviceInfo *info;
+                if (l) {
                         GUPnPDeviceProxy *proxy;
 
-                        info = GUPNP_DEVICE_INFO (l->data);
-
-                        if (strcmp (udn,
-                                    gupnp_device_info_get_udn (info)) != 0) {
-                                l = l->next;
-
-                                continue;
-                        }
-
                         /* Remove proxy */
-                        proxy = GUPNP_DEVICE_PROXY (info);
+                        proxy = GUPNP_DEVICE_PROXY (l->data);
 
                         cur_l = l;
                         l = l->next;
@@ -766,7 +844,8 @@ gupnp_control_point_get_property (GObject    *object,
 
         switch (property_id) {
         case PROP_RESOURCE_FACTORY:
-                g_value_set_object (value, control_point->priv->factory);
+                g_value_set_object (value,
+                                    gupnp_control_point_get_resource_factory (control_point));
                 break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -916,13 +995,13 @@ GUPnPControlPoint *
 gupnp_control_point_new (GUPnPContext *context,
                          const char   *target)
 {
-        GUPnPResourceFactory *factory;
+        g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
+        g_return_val_if_fail (target, NULL);
 
-        factory = gupnp_resource_factory_get_default ();
-
-        return gupnp_control_point_new_full (context,
-                                             factory,
-                                             target);
+        return g_object_new (GUPNP_TYPE_CONTROL_POINT,
+                             "client", context,
+                             "target", target,
+                             NULL);
 }
 
 /**
@@ -946,7 +1025,8 @@ gupnp_control_point_new_full (GUPnPContext         *context,
                               const char           *target)
 {
         g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
-        g_return_val_if_fail (GUPNP_IS_RESOURCE_FACTORY (factory), NULL);
+        g_return_val_if_fail (factory == NULL ||
+                              GUPNP_IS_RESOURCE_FACTORY (factory), NULL);
         g_return_val_if_fail (target, NULL);
 
         return g_object_new (GUPNP_TYPE_CONTROL_POINT,
@@ -1024,5 +1104,8 @@ gupnp_control_point_get_resource_factory (GUPnPControlPoint *control_point)
 {
         g_return_val_if_fail (GUPNP_IS_CONTROL_POINT (control_point), NULL);
 
-        return control_point->priv->factory;
+        if (control_point->priv->factory)
+                  return control_point->priv->factory;
+
+        return gupnp_resource_factory_get_default ();
 }
