@@ -40,18 +40,20 @@ G_DEFINE_TYPE (GUPnPRootDevice,
                GUPNP_TYPE_DEVICE);
 
 struct _GUPnPRootDevicePrivate {
-        xmlDoc *description_doc;
-        gboolean own_description_doc;
+        GUPnPXMLDoc *description_doc;
 
         GSSDPResourceGroup *group;
 
-        char *relative_location;
+        char  *description_path;
+        char  *description_dir;
+        char  *relative_location;
 };
 
 enum {
         PROP_0,
         PROP_DESCRIPTION_DOC,
-        PROP_RELATIVE_LOCATION,
+        PROP_DESCRIPTION_PATH,
+        PROP_DESCRIPTION_DIR,
         PROP_AVAILABLE
 };
 
@@ -63,11 +65,10 @@ gupnp_root_device_finalize (GObject *object)
 
         device = GUPNP_ROOT_DEVICE (object);
 
+        g_free (device->priv->description_path);
+        g_free (device->priv->description_dir);
         g_free (device->priv->relative_location);
 
-        if (device->priv->own_description_doc)
-                xmlFreeDoc (device->priv->description_doc);
-
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_root_device_parent_class);
         object_class->finalize (object);
@@ -86,6 +87,11 @@ gupnp_root_device_dispose (GObject *object)
                 device->priv->group = NULL;
         }
 
+        if (device->priv->description_doc) {
+                g_object_unref (device->priv->description_doc);
+                device->priv->description_doc = NULL;
+        }
+
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_root_device_parent_class);
         object_class->dispose (object);
@@ -97,8 +103,6 @@ gupnp_root_device_init (GUPnPRootDevice *device)
         device->priv = G_TYPE_INSTANCE_GET_PRIVATE (device,
                                                     GUPNP_TYPE_ROOT_DEVICE,
                                                     GUPnPRootDevicePrivate);
-
-        device->priv->own_description_doc = FALSE;
 }
 
 static void
@@ -113,10 +117,13 @@ gupnp_root_device_set_property (GObject      *object,
 
         switch (property_id) {
         case PROP_DESCRIPTION_DOC:
-                device->priv->description_doc = g_value_get_pointer (value);
+                device->priv->description_doc = g_value_dup_object (value);
                 break;
-        case PROP_RELATIVE_LOCATION:
-                device->priv->relative_location = g_value_dup_string (value);
+        case PROP_DESCRIPTION_PATH:
+                device->priv->description_path = g_value_dup_string (value);
+                break;
+        case PROP_DESCRIPTION_DIR:
+                device->priv->description_dir = g_value_dup_string (value);
                 break;
         case PROP_AVAILABLE:
                 gupnp_root_device_set_available
@@ -139,9 +146,13 @@ gupnp_root_device_get_property (GObject    *object,
         device = GUPNP_ROOT_DEVICE (object);
 
         switch (property_id) {
-        case PROP_RELATIVE_LOCATION:
+        case PROP_DESCRIPTION_PATH:
                 g_value_set_string (value,
-                                    device->priv->relative_location);
+                                    device->priv->description_path);
+                break;
+        case PROP_DESCRIPTION_DIR:
+                g_value_set_string (value,
+                                    device->priv->description_dir);
                 break;
         case PROP_AVAILABLE:
                 g_value_set_boolean (value,
@@ -153,57 +164,6 @@ gupnp_root_device_get_property (GObject    *object,
         }
 }
 
-/* Adds @udn::@resource_type to @group, together with all earlier versions
- * of @resource_type.
- *
- * Note: @resource_type is modified in this function. */
-static void
-add_resource_with_earlier_versions (GSSDPResourceGroup *group,
-                                    const char         *udn,
-                                    const char         *resource_type,
-                                    const char         *location)
-{
-        char *colon, *usn;
-        int version, i;
-
-        /* Add specified version */
-        usn = g_strdup_printf ("%s::%s", (char *) udn, (char *) resource_type);
-        gssdp_resource_group_add_resource_simple (group,
-                                                  (const char *) resource_type,
-                                                  usn,
-                                                  location);
-        g_free (usn);
-
-        /* Try to parse version */
-        colon = strrchr (resource_type, ':');
-        if (G_UNLIKELY (*colon == 0))
-                return;
-
-        colon += 1;
-        if (G_UNLIKELY (*colon == 0))
-                return;
-
-        version = atoi (colon);
-        
-        /* Add earlier versions */
-        *colon = 0;
-
-        for (i = 1; i < version; i++) {
-                char *type;
-
-                type = g_strdup_printf ("%s%d", (char *) resource_type, i);
-                usn = g_strdup_printf ("%s::%s", (char *) udn, (char *) type);
-
-                gssdp_resource_group_add_resource_simple (group,
-                                                          type,
-                                                          usn,
-                                                          location);
-
-                g_free (usn);
-                g_free (type);
-        }
-}
-
 static void
 fill_resource_group (xmlNode            *element,
                      const char         *location,
@@ -211,6 +171,7 @@ fill_resource_group (xmlNode            *element,
 {
         xmlNode *child;
         xmlChar *udn, *device_type;
+        char *usn;
 
         /* Add device */
         udn = xml_util_get_child_element_content (element, "UDN");
@@ -233,10 +194,12 @@ fill_resource_group (xmlNode            *element,
                                                   (const char *) udn,
                                                   location);
 
-        add_resource_with_earlier_versions (group,
-                                            (const char *) udn,
-                                            (const char *) device_type,
-                                            location);
+        usn = g_strdup_printf ("%s::%s", (char *) udn, (char *) device_type);
+        gssdp_resource_group_add_resource_simple (group,
+                                                  (const char *) device_type,
+                                                  (const char *) usn,
+                                                  location);
+        g_free (usn);
 
         xmlFree (device_type);
 
@@ -256,11 +219,15 @@ fill_resource_group (xmlNode            *element,
                         if (!service_type)
                                 continue;
 
-                        add_resource_with_earlier_versions
-                                (group,
-                                 (const char *) udn,
-                                 (const char *) service_type,
-                                 location);
+                        usn = g_strdup_printf ("%s::%s",
+                                               (char *) udn,
+                                               (char *) service_type);
+                        gssdp_resource_group_add_resource_simple
+                                                (group,
+                                                 (const char *) service_type,
+                                                 (const char *) usn,
+                                                 location);
+                        g_free (usn);
 
                         xmlFree (service_type);
                 }
@@ -280,47 +247,20 @@ fill_resource_group (xmlNode            *element,
         }
 }
 
-/* Download and parse @location as an XML document, synchronously. */
-static xmlDoc *
-download_and_parse (GUPnPContext *context,
-                    const char   *location)
+/* Load and parse @description_path as an XML document, synchronously. */
+static GUPnPXMLDoc *
+load_and_parse (const char *description_path)
 {
-        xmlDoc *description_doc;
-        SoupSession *session;
-        SoupMessage *msg;
-        int status;
-
-        session = gupnp_context_get_session (context);
-
-        msg = soup_message_new (SOUP_METHOD_GET, location);
-        if (msg == NULL) {
-                g_warning ("Invalid URL: %s", location);
-
-                return NULL;
-        }
-
-        http_request_set_user_agent (msg);
-        http_request_set_accept_language (msg);
-
-        status = soup_session_send_message (session, msg);
-        if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
-                g_warning ("Unable to download description document "
-                           "from %s", location);
-
-                g_object_unref (msg);
-
-                return NULL;
-        }
-
-        description_doc = xmlRecoverMemory (msg->response_body->data,
-                                            msg->response_body->length); 
-
-        g_object_unref (msg);
+        GUPnPXMLDoc *description_doc;
+        GError *error = NULL;
 
+        /* We don't, so load and parse it */
+        description_doc = gupnp_xml_doc_new_from_path (description_path,
+                                                       &error);
         if (description_doc == NULL) {
-                g_warning ("Failed to parse %s\n", location);
+                g_critical ("Error loading description: %s\n", error->message);
 
-                return NULL;
+                g_error_free (error);
         }
 
         return description_doc;
@@ -335,22 +275,22 @@ gupnp_root_device_constructor (GType                  type,
         GObject *object;
         GUPnPRootDevice *device;
         GUPnPContext *context;
-        const char *relative_location, *udn;
-        gboolean own_description_doc;
-        char *location, *usn;
+        const char *description_path, *description_dir, *udn;
+        char *desc_path, *location, *usn, *relative_location;
         int i;
-        xmlDoc *description_doc;
+        GUPnPXMLDoc *description_doc;
         xmlNode *root_element, *element;
         SoupURI *url_base;
 
         object = NULL;
-        own_description_doc = FALSE;
+        location = NULL;
 
-        /* Get 'description-doc', 'context' and 'relative-location' property
-         * values */
+        /* Get 'description-doc', 'context', 'description-dir' and
+         * 'description-path' property values */
         description_doc   = NULL;
         context           = NULL;
-        relative_location = NULL;
+        description_path  = NULL;
+        description_dir   = NULL;
 
         for (i = 0; i < n_construct_params; i++) {
                 const char *par_name;
@@ -359,7 +299,7 @@ gupnp_root_device_constructor (GType                  type,
 
                 if (strcmp (par_name, "description-doc") == 0) {
                         description_doc =
-                                g_value_get_pointer (construct_params[i].value);
+                                g_value_get_object (construct_params[i].value);
 
                         continue;
                 } 
@@ -371,8 +311,15 @@ gupnp_root_device_constructor (GType                  type,
                         continue;
                 }
 
-                if (strcmp (par_name, "relative-location") == 0) {
-                        relative_location =
+                if (strcmp (par_name, "description-path") == 0) {
+                        description_path =
+                                g_value_get_string (construct_params[i].value);
+
+                        continue;
+                }
+
+                if (strcmp (par_name, "description-dir") == 0) {
+                        description_dir =
                                 g_value_get_string (construct_params[i].value);
 
                         continue;
@@ -385,31 +332,35 @@ gupnp_root_device_constructor (GType                  type,
                 return NULL;
         }
 
-        if (!relative_location) {
-                g_warning ("No relative location specified.");
+        if (!description_path) {
+                g_warning ("Path to description document not specified.");
 
                 return NULL;
         }
 
+        if (!description_dir) {
+                g_warning ("Path to description directory not specified.");
 
-        /* Generate full location */
-        location = g_build_path ("/",
-                                 _gupnp_context_get_server_url (context),
-                                 relative_location,
-                                 NULL);
+                return NULL;
+        }
+
+        if (g_path_is_absolute (description_path))
+                desc_path = g_strdup (description_path);
+        else
+                desc_path = g_build_filename (description_dir,
+                                              description_path,
+                                              NULL);
 
         /* Check whether we have a parsed description document */
         if (!description_doc) {
-                /* We don't, so download and parse it */
-                description_doc = download_and_parse (context, location);
+                /* We don't, so load and parse it */
+                description_doc = load_and_parse (desc_path);
                 if (description_doc == NULL)
                         goto DONE;
-                
-                own_description_doc = TRUE;
         }
 
         /* Find correct element */
-        root_element = xml_util_get_element ((xmlNode *) description_doc,
+        root_element = xml_util_get_element ((xmlNode *) description_doc->doc,
                                              "root",
                                              NULL);
         if (!root_element) {
@@ -441,8 +392,8 @@ gupnp_root_device_constructor (GType                  type,
                 }
 
                 if (strcmp (par_name, "description-doc") == 0) {
-                        g_value_set_pointer (construct_params[i].value,
-                                             description_doc);
+                        g_value_set_object (construct_params[i].value,
+                                            description_doc);
 
                         continue;
                 }
@@ -456,6 +407,27 @@ gupnp_root_device_constructor (GType                  type,
                                             construct_params);
         device = GUPNP_ROOT_DEVICE (object);
 
+        /* Generate location relative to HTTP root */
+        device->priv->relative_location = g_strdup_printf ("RootDevice%p.xml",                                                             object);
+
+        relative_location = g_strjoin (NULL,
+                                       "/",
+                                       device->priv->relative_location,
+                                       NULL);
+
+        /* Host the description file and dir */
+        gupnp_context_host_path (context,
+                                 desc_path,
+                                 relative_location);
+        gupnp_context_host_path (context, device->priv->description_dir, "");
+
+        /* Generate full location */
+        location = g_strjoin (NULL,
+                              _gupnp_context_get_server_url (context),
+                              relative_location,
+                              NULL);
+        g_free (relative_location);
+
         /* Save the URL base, if any */
         url_base = xml_util_get_child_element_content_uri (root_element,
                                                            "URLBase",
@@ -471,8 +443,6 @@ gupnp_root_device_constructor (GType                  type,
 
 	soup_uri_free (url_base);
 
-        device->priv->own_description_doc = own_description_doc;
-
         /* Create resource group */
         device->priv->group = gssdp_resource_group_new (GSSDP_CLIENT (context));
 
@@ -512,31 +482,52 @@ gupnp_root_device_class_init (GUPnPRootDeviceClass *klass)
         /**
          * GUPnPRootDevice:description-doc
          *
-         * Pointer to description document. Constructor property.
+         * Device description document. Constructor property.
          **/
         g_object_class_install_property
                 (object_class,
                  PROP_DESCRIPTION_DOC,
-                 g_param_spec_pointer ("description-doc",
-                                       "Description document",
-                                       "Pointer to description document",
-                                       G_PARAM_WRITABLE |
-                                       G_PARAM_CONSTRUCT_ONLY |
-                                       G_PARAM_STATIC_NAME |
-                                       G_PARAM_STATIC_NICK |
-                                       G_PARAM_STATIC_BLURB));
+                 g_param_spec_object ("description-doc",
+                                      "Description document",
+                                      "Device description document",
+                                      GUPNP_TYPE_XML_DOC,
+                                      G_PARAM_WRITABLE |
+                                      G_PARAM_CONSTRUCT_ONLY |
+                                      G_PARAM_STATIC_NAME |
+                                      G_PARAM_STATIC_NICK |
+                                      G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPRootDevice:relative-location
+         * GUPnPRootDevice:description-path
          *
-         * Relative location.
+         * The path to device description document. This could either be an
+         * absolute path or path relative to GUPnPRootDevice:description-dir.
          **/
         g_object_class_install_property
                 (object_class,
-                 PROP_RELATIVE_LOCATION,
-                 g_param_spec_string ("relative-location",
-                                      "Relative location",
-                                      "Relative location",
+                 PROP_DESCRIPTION_PATH,
+                 g_param_spec_string ("description-path",
+                                      "Description Path",
+                                      "The path to device descrition document",
+                                      NULL,
+                                      G_PARAM_READWRITE |
+                                      G_PARAM_CONSTRUCT_ONLY |
+                                      G_PARAM_STATIC_NAME |
+                                      G_PARAM_STATIC_NICK |
+                                      G_PARAM_STATIC_BLURB));
+
+        /**
+         * GUPnPRootDevice:description-dir
+         *
+         * The path to directory where description documents are provided.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_DESCRIPTION_DIR,
+                 g_param_spec_string ("description-dir",
+                                      "Description Directory",
+                                      "The path to directory where "
+                                      "description documents are provided",
                                       NULL,
                                       G_PARAM_READWRITE |
                                       G_PARAM_CONSTRUCT_ONLY |
@@ -565,17 +556,19 @@ gupnp_root_device_class_init (GUPnPRootDeviceClass *klass)
 /**
  * gupnp_root_device_new
  * @context: The #GUPnPContext
- * @relative_location: Location of the description file for this device,
- * relative to the HTTP root
+ * @description_path: Path to device description document. This could either
+ * be an absolute path or path relative to @description_dir.
+ * @description_dir: Path to directory where description documents are provided.
  *
- * Create a new #GUPnPRootDevice object, automatically downloading and
- * parsing @relative_location.
+ * Create a new #GUPnPRootDevice object, automatically loading and parsing
+ * device description document from @description_path.
  *
  * Return value: A new @GUPnPRootDevice object.
  **/
 GUPnPRootDevice *
 gupnp_root_device_new (GUPnPContext *context,
-                       const char   *relative_location)
+                       const char   *description_path,
+                       const char   *description_dir)
 {
         GUPnPResourceFactory *factory;
 
@@ -584,27 +577,31 @@ gupnp_root_device_new (GUPnPContext *context,
         return gupnp_root_device_new_full (context,
                                            factory,
                                            NULL,
-                                           relative_location);
+                                           description_path,
+                                           description_dir);
 }
 
 /**
  * gupnp_root_device_new_full
  * @context: A #GUPnPContext
  * @factory: A #GUPnPResourceFactory
- * @description_doc: Pointer to the device description document, or %NULL
- * @relative_location: Location of the description file for this device,
- * relative to the HTTP root
+ * @description_doc: Device description document, or %NULL
+ * @description_path: Path to device description document. This could either
+ * be an absolute path or path relative to @description_dir.
+ * @description_dir: Path to directory where description documents are provided.
  *
- * Create a new #GUPnPRootDevice, automatically downloading and parsing
- * @relative_location if @description_doc is %NULL.
+ * Create a new #GUPnPRootDevice, automatically loading and parsing
+ * device description document from @description_path if @description_doc is
+ * %NULL.
  *
  * Return value: A new #GUPnPRootDevice object.
  **/
 GUPnPRootDevice *
 gupnp_root_device_new_full (GUPnPContext         *context,
                             GUPnPResourceFactory *factory,
-                            xmlDoc               *description_doc,
-                            const char           *relative_location)
+                            GUPnPXMLDoc          *description_doc,
+                            const char           *description_path,
+                            const char           *description_dir)
 {
         g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
         g_return_val_if_fail (GUPNP_IS_RESOURCE_FACTORY (factory), NULL);
@@ -614,7 +611,8 @@ gupnp_root_device_new_full (GUPnPContext         *context,
                              "resource-factory", factory,
                              "root-device", NULL,
                              "description-doc", description_doc,
-                             "relative-location", relative_location,
+                             "description-path", description_path,
+                             "description-dir", description_dir,
                              NULL);
 }
 
@@ -669,3 +667,36 @@ gupnp_root_device_get_relative_location (GUPnPRootDevice *root_device)
 
         return root_device->priv->relative_location;
 }
+
+/**
+ * gupnp_root_device_get_description_path
+ * @root_device: A #GUPnPRootDevice
+ *
+ * Get the path to the device description document of @root_device.
+ *
+ * Return value: The path to device description document of @root_device.
+ **/
+const char *
+gupnp_root_device_get_description_path (GUPnPRootDevice *root_device)
+{
+        g_return_val_if_fail (GUPNP_IS_ROOT_DEVICE (root_device), NULL);
+
+        return root_device->priv->description_path;
+}
+
+/**
+ * gupnp_root_device_get_description_dir
+ * @root_device: A #GUPnPRootDevice
+ *
+ * Get the path to the directory containing description documents related to
+ * @root_device.
+ *
+ * Return value: The path to description document directory of @root_device.
+ **/
+const char *
+gupnp_root_device_get_description_dir (GUPnPRootDevice *root_device)
+{
+        g_return_val_if_fail (GUPNP_IS_ROOT_DEVICE (root_device), NULL);
+
+        return root_device->priv->description_dir;
+}
