@@ -1446,39 +1446,14 @@ server_handler (SoupServer        *soup_server,
 
         hdr = soup_message_headers_get (msg->request_headers, "SID");
         if (hdr == NULL ||
-            (proxy->priv->sid && (strcmp (hdr, proxy->priv->sid) != 0))) {
+            (proxy->priv->sid == NULL ||
+             strcmp (hdr, proxy->priv->sid) != 0)) {
                 /* No SID or not ours */
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
 
                 return;
         }
 
-        /* We do not error out if proxy->priv->sid is NIL as the subscription
-         * response may not have been processed yet. */
-#if 0
-        if (!proxy->priv->sid) {
-                GUPnPContext *context;
-                GMainContext *main_context;
-
-                context = gupnp_service_info_get_context
-                                        (GUPNP_SERVICE_INFO (proxy));
-                main_context = gssdp_client_get_main_context
-                                        (GSSDP_CLIENT (context));
-
-                /* Wait. Perhaps the subscription response has not yet
-                 * been processed. */
-                g_main_context_iteration (main_context, FALSE);
-
-                if (!proxy->priv->sid || strcmp (hdr, proxy->priv->sid) != 0) {
-                        /* Really not our SID */
-                        soup_message_set_status
-                                (msg, SOUP_STATUS_PRECONDITION_FAILED);
-
-                        return;
-                }
-        }
-#endif
-
         hdr = soup_message_headers_get (msg->request_headers, "SEQ");
         if (hdr == NULL) {
                 /* No SEQ header */
@@ -1602,7 +1577,7 @@ subscription_expire (gpointer user_data)
 
         g_free (sub_url);
 
-        g_assert (msg != NULL);
+        g_return_val_if_fail (msg != NULL, FALSE);
 
         /* Add headers */
         soup_message_headers_append (msg->request_headers,
@@ -1670,7 +1645,7 @@ subscribe_got_response (SoupSession       *session,
                                          GUPNP_EVENTING_ERROR_SUBSCRIPTION_LOST,
                                          "No SID in SUBSCRIBE response");
 
-                        goto ERROR;
+                        goto hdr_err;
                 }
 
                 proxy->priv->sid = g_strdup (hdr);
@@ -1730,7 +1705,7 @@ subscribe_got_response (SoupSession       *session,
                                  GUPNP_EVENTING_ERROR_SUBSCRIPTION_FAILED,
                                  msg->reason_phrase);
 
- ERROR:
+hdr_err:
                 proxy->priv->subscribed = FALSE;
 
                 g_object_notify (G_OBJECT (proxy), "subscribed");
@@ -1855,11 +1830,15 @@ unsubscribe (GUPnPServiceProxy *proxy)
         SoupServer *server;
         char *sub_url;
 
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (proxy));
+
+        /* Remove server handler */
+        server = gupnp_context_get_server (context);
+        soup_server_remove_handler (server, proxy->priv->path);
+
         if (proxy->priv->sid == NULL)
                 return; /* No SID: nothing to unsubscribe */
 
-        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (proxy));
-
         /* Create unsubscription message */
         sub_url = gupnp_service_info_get_event_subscription_url
                                                 (GUPNP_SERVICE_INFO (proxy));
@@ -1868,17 +1847,17 @@ unsubscribe (GUPnPServiceProxy *proxy)
 
         g_free (sub_url);
 
-        g_assert (msg != NULL);
+        if (msg != NULL) {
+                /* Add headers */
+                soup_message_headers_append (msg->request_headers,
+                                             "SID",
+                                             proxy->priv->sid);
 
-        /* Add headers */
-        soup_message_headers_append (msg->request_headers,
-                                     "SID",
-                                     proxy->priv->sid);
+                /* And queue it */
+                session = gupnp_context_get_session (context);
 
-        /* And queue it */
-        session = gupnp_context_get_session (context);
-
-        soup_session_queue_message (session, msg, NULL, NULL);
+                soup_session_queue_message (session, msg, NULL, NULL);
+        }
 
         /* Reset SID */
         g_free (proxy->priv->sid);
@@ -1892,10 +1871,6 @@ unsubscribe (GUPnPServiceProxy *proxy)
                 g_source_destroy (proxy->priv->subscription_timeout_src);
                 proxy->priv->subscription_timeout_src = NULL;
         }
-
-        /* Remove server handler */
-        server = gupnp_context_get_server (context);
-        soup_server_remove_handler (server, proxy->priv->path);
 }
 
 /**
