@@ -24,51 +24,6 @@
 #include "gvalue-util.h"
 #include "xml-util.h"
 
-G_DEFINE_TYPE (XmlDocWrapper,
-               xml_doc_wrapper,
-               G_TYPE_INITIALLY_UNOWNED);
-
-static void
-xml_doc_wrapper_init (XmlDocWrapper *wrapper)
-{
-        /* Empty */
-}
-
-static void
-xml_doc_wrapper_finalize (GObject *object)
-{
-        XmlDocWrapper *wrapper;
-
-        wrapper = XML_DOC_WRAPPER (object);
-
-        xmlFreeDoc (wrapper->doc);
-}
-
-static void
-xml_doc_wrapper_class_init (XmlDocWrapperClass *klass)
-{
-        GObjectClass *object_class;
-
-        object_class = G_OBJECT_CLASS (klass);
-
-        object_class->finalize = xml_doc_wrapper_finalize;
-}
-
-/* Takes ownership of @doc */
-XmlDocWrapper *
-xml_doc_wrapper_new (xmlDoc *doc)
-{
-        XmlDocWrapper *wrapper;
-
-        g_return_val_if_fail (doc != NULL, NULL);
-
-        wrapper = g_object_new (TYPE_XML_DOC_WRAPPER, NULL);
-
-        wrapper->doc = doc;
-
-        return wrapper;
-}
-
 /* libxml DOM interface helpers */
 xmlNode *
 xml_util_get_element (xmlNode *node,
@@ -149,13 +104,13 @@ xml_util_get_child_element_content_glib (xmlNode    *node,
         return copy;
 }
 
-SoupUri *
+SoupURI *
 xml_util_get_child_element_content_uri (xmlNode    *node,
                                         const char *child_name,
-                                        SoupUri    *base)
+                                        SoupURI    *base)
 {
         xmlChar *content;
-        SoupUri *uri;
+        SoupURI *uri;
 
         content = xml_util_get_child_element_content (node, child_name);
         if (!content)
@@ -171,9 +126,9 @@ xml_util_get_child_element_content_uri (xmlNode    *node,
 char *
 xml_util_get_child_element_content_url (xmlNode    *node,
                                         const char *child_name,
-                                        SoupUri    *base)
+                                        SoupURI    *base)
 {
-        SoupUri *uri;
+        SoupURI *uri;
         char *url;
 
         uri = xml_util_get_child_element_content_uri (node, child_name, base);
@@ -206,6 +161,26 @@ xml_util_get_attribute_contents (xmlNode    *node,
                 return NULL;
 }
 
+/**
+ * xml_util_real_node:
+ * @node: an %xmlNodePtr
+ *
+ * Finds the first "real" node (ie, not a comment or whitespace) at or
+ * after @node at its level in the tree.
+ *
+ * Return: a node, or %NULL
+ *
+ * (Taken from libsoup)
+ **/
+xmlNode *
+xml_util_real_node (xmlNode *node)
+{
+	while (node && (node->type == XML_COMMENT_NODE ||
+			xmlIsBlankNode (node)))
+		node = node->next;
+	return node;
+}
+
 /* XML string creation helpers */
 
 #define INITIAL_XML_STR_SIZE 100 /* Initial xml string size in bytes */
